from transformers import PreTrainedModel
from transformers.models.marian.modeling_marian import MarianEncoder, MarianDecoder
import torch
import torch.nn as nn
from typing import List, Union, Optional, Tuple, Dict
from transformers.modeling_outputs import BaseModelOutput, Seq2SeqModelOutput, Seq2SeqLMOutput
from contextlib import contextmanager
from dataclasses import dataclass
from marian import MarianOTConfig

class MarianOTPreTrainedModel(PreTrainedModel):
    config_class = MarianOTConfig
    base_model_prefix = "marian_ot"
    supports_gradient_checkpointing = True

    def _init_weights(self, module: Union[nn.Linear, nn.Embedding]):
        std = self.config.init_std
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()

    def _set_gradient_checkpointing(self, module, value=False):
        if isinstance(module, (MarianDecoder, MarianEncoder)):
            module.gradient_checkpointing = value

    @property
    def dummy_inputs(self):
        pad_token = self.config.pad_token_id
        input_ids = torch.tensor([[0, 6, 10, 4, 2], [0, 8, 12, 2, pad_token]], device=self.device)
        dummy_inputs = {
            "attention_mask": input_ids.ne(pad_token),
            "input_ids": input_ids,
            "decoder_input_ids": input_ids,
        }
        return dummy_inputs

@dataclass
class CombineSeq2SeqLMOutput:
    tgt: Seq2SeqLMOutput
    tsl: Seq2SeqLMOutput

class MarianOT(MarianOTPreTrainedModel):
    def __init__(self, config: MarianOTConfig):
        super().__init__(config)

        padding_idx, vocab_size = config.pad_token_id, config.vocab_size
        # We always use self.shared for token embeddings to ensure compatibility with all marian models
        self.shared = nn.Embedding(vocab_size, config.d_model, padding_idx)
        if self.config.share_encoder_decoder_embeddings:
            encoder_embed_tokens = decoder_embed_tokens = self.shared
        else:
            # Since the embeddings are not shared, deepcopy the embeddings here for encoder
            # and decoder to make sure they are not tied.
            encoder_embed_tokens = copy.deepcopy(self.shared)
            decoder_embed_tokens = copy.deepcopy(self.shared)
            self.shared = None

        self.encoder = MarianEncoder(config, encoder_embed_tokens)
        self.tgt_decoder = MarianDecoder(config, decoder_embed_tokens)
        self.tgt_lm_head = nn.Linear(config.d_model, config.vocab_size, bias=True)
        self.tsl_decoder = MarianDecoder(config, decoder_embed_tokens)
        self.tsl_lm_head = nn.Linear(config.d_model, config.vocab_size, bias=True)
        
        self.translate_self_flag = False

        # Initialize weights and apply final processing
        self.post_init()

    @contextmanager
    def translate_self(self):
        self.translate_self_flag = True
        yield
        self.translate_self_flag = False

    def get_input_embeddings(self):
        # This will return shared embeddings if they are shared else specific to encoder.
        return self.get_encoder().get_input_embeddings()

    def set_input_embeddings(self, value):
        if self.config.share_encoder_decoder_embeddings:
            self.shared = value
            self.encoder.embed_tokens = self.shared
            self.tsl_decoder.embed_tokens = self.shared
            self.tgt_decoder.embed_tokens = self.shared
        else:  # if not shared only set encoder embeedings
            self.encoder.embed_tokens = value

    def get_decoder_input_embeddings(self):
        if self.config.share_encoder_decoder_embeddings:
            raise ValueError(
                "`get_decoder_input_embeddings` should not be called if `config.share_encoder_decoder_embeddings` "
                "is `True`. Please use `get_input_embeddings` instead."
            )
        return self.get_decoder().get_input_embeddings()

    def set_decoder_input_embeddings(self, value):
        if self.config.share_encoder_decoder_embeddings:
            raise ValueError(
                "`config.share_encoder_decoder_embeddings` is set to `True` meaning the decoder input embeddings "
                "are shared with the encoder. In order to set the decoder input embeddings, you should simply set "
                "the encoder input embeddings by calling `set_input_embeddings` with the appropriate embeddings."
            )
        self.tsl_decoder.embed_tokens = value
        self.tgt_decoder.embed_tokens = value

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        if self.translate_self_flag:
            return self.tgt_decoder
        else:
            return self.tsl_decoder

    def resize_decoder_token_embeddings(self, new_num_tokens: int) -> nn.Embedding:
        if self.config.share_encoder_decoder_embeddings:
            raise ValueError(
                "`resize_decoder_token_embeddings` should not be called if `config.share_encoder_decoder_embeddings` "
                "is `True`. Please use `resize_token_embeddings` instead."
            )

        old_embeddings = self.get_decoder_input_embeddings()
        new_embeddings = self._get_resized_embeddings(old_embeddings, new_num_tokens)
        self.set_decoder_input_embeddings(new_embeddings)

        model_embeds = self.get_decoder_input_embeddings()

        if new_num_tokens is None:
            return model_embeds

        # Update base model and current model config
        self.config.decoder_vocab_size = new_num_tokens

        # Tie weights again if needed
        self.tie_weights()

        return model_embeds

    def forward_pass(
        self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        tsl_decoder_input_ids: Optional[torch.LongTensor] = None,
        tgt_decoder_input_ids: Optional[torch.LongTensor] = None,
        tsl_decoder_attention_mask: Optional[torch.Tensor] = None,
        tgt_decoder_attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        tsl_decoder_head_mask: Optional[torch.Tensor] = None,
        tgt_decoder_head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        encoder_outputs: Optional[Union[Tuple[torch.Tensor], BaseModelOutput]] = None,
        past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        decoder_inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> CombineSeq2SeqLMOutput:
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if encoder_outputs is None:
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                head_mask=head_mask,
                inputs_embeds=inputs_embeds,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
            )
        # If the user passed a tuple for encoder_outputs, we wrap it in a BaseModelOutput when return_dict=True
        elif return_dict and not isinstance(encoder_outputs, BaseModelOutput):
            encoder_outputs = BaseModelOutput(
                last_hidden_state=encoder_outputs[0],
                hidden_states=encoder_outputs[1] if len(encoder_outputs) > 1 else None,
                attentions=encoder_outputs[2] if len(encoder_outputs) > 2 else None,
            )

        decoder_outputs_tsl = self.tsl_decoder(
            input_ids=tsl_decoder_input_ids,
            attention_mask=tsl_decoder_attention_mask,
            encoder_hidden_states=encoder_outputs[0],
            encoder_attention_mask=attention_mask,
            head_mask=tsl_decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=decoder_inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        decoder_outputs_tgt = self.tgt_decoder(
            input_ids=tgt_decoder_input_ids,
            attention_mask=tgt_decoder_attention_mask,
            encoder_hidden_states=encoder_outputs[0],
            encoder_attention_mask=attention_mask,
            head_mask=tgt_decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=decoder_inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        tsl_outputs = Seq2SeqModelOutput(
            last_hidden_state=decoder_outputs_tsl.last_hidden_state,
            past_key_values=decoder_outputs_tsl.past_key_values,
            decoder_hidden_states=decoder_outputs_tsl.hidden_states,
            decoder_attentions=decoder_outputs_tsl.attentions,
            cross_attentions=decoder_outputs_tsl.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )

        tgt_outputs = Seq2SeqModelOutput(
            last_hidden_state=decoder_outputs_tgt.last_hidden_state,
            past_key_values=decoder_outputs_tgt.past_key_values,
            decoder_hidden_states=decoder_outputs_tgt.hidden_states,
            decoder_attentions=decoder_outputs_tgt.attentions,
            cross_attentions=decoder_outputs_tgt.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )

        tgt_lm_logits = self.tgt_lm_head(tgt_outputs[0])
        tsl_lm_logits = self.tsl_lm_head(tsl_outputs[0])

        tgt_lm_output = Seq2SeqLMOutput(
            loss=None,
            logits=tgt_lm_logits,
            past_key_values=tgt_outputs.past_key_values,
            decoder_hidden_states=tgt_outputs.decoder_hidden_states,
            decoder_attentions=tgt_outputs.decoder_attentions,
            cross_attentions=tgt_outputs.cross_attentions,
            encoder_last_hidden_state=tgt_outputs.encoder_last_hidden_state,
            encoder_hidden_states=tgt_outputs.encoder_hidden_states,
            encoder_attentions=tgt_outputs.encoder_attentions,
        )

        tsl_lm_output = Seq2SeqLMOutput(
            loss=None,
            logits=tsl_lm_logits,
            past_key_values=tsl_outputs.past_key_values,
            decoder_hidden_states=tsl_outputs.decoder_hidden_states,
            decoder_attentions=tsl_outputs.decoder_attentions,
            cross_attentions=tsl_outputs.cross_attentions,
            encoder_last_hidden_state=tsl_outputs.encoder_last_hidden_state,
            encoder_hidden_states=tsl_outputs.encoder_hidden_states,
            encoder_attentions=tsl_outputs.encoder_attentions,
        )

        return CombineSeq2SeqLMOutput(
            tgt=tgt_lm_output,
            tsl=tsl_lm_output
        )

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        decoder_input_ids: Optional[torch.LongTensor] = None,
        decoder_attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        decoder_head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        encoder_outputs: Optional[Union[Tuple[torch.Tensor], BaseModelOutput]] = None,
        past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        decoder_inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Seq2SeqLMOutput:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for computing the masked language modeling loss. Indices should either be in `[0, ...,
            config.vocab_size]` or -100 (see `input_ids` docstring). Tokens with indices set to `-100` are ignored
            (masked), the loss is only computed for the tokens with labels in `[0, ..., config.vocab_size]`.
        Returns:
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if labels is not None:
            if use_cache:
                logger.warning("The `use_cache` argument is changed to `False` since `labels` is provided.")
            use_cache = False
            if decoder_input_ids is None and decoder_inputs_embeds is None:
                decoder_input_ids = shift_tokens_right(
                    labels, self.config.pad_token_id, self.config.decoder_start_token_id
                )

        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if encoder_outputs is None:
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                head_mask=head_mask,
                inputs_embeds=inputs_embeds,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
            )
        # If the user passed a tuple for encoder_outputs, we wrap it in a BaseModelOutput when return_dict=True
        elif return_dict and not isinstance(encoder_outputs, BaseModelOutput):
            encoder_outputs = BaseModelOutput(
                last_hidden_state=encoder_outputs[0],
                hidden_states=encoder_outputs[1] if len(encoder_outputs) > 1 else None,
                attentions=encoder_outputs[2] if len(encoder_outputs) > 2 else None,
            )

        # decoder outputs consists of (dec_features, past_key_value, dec_hidden, dec_attn)
        if self.translate_self_flag:
            decoder_fn = self.tgt_decoder
            lm_fn = self.tgt_lm_head
        else:
            decoder_fn = self.tsl_decoder
            lm_fn = self.tsl_lm_head

        decoder_outputs = decoder_fn(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            encoder_hidden_states=encoder_outputs[0],
            encoder_attention_mask=attention_mask,
            head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=decoder_inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        outputs = Seq2SeqModelOutput(
            last_hidden_state=decoder_outputs.last_hidden_state,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )
        
        lm_logits = lm_fn(outputs[0])

        masked_lm_loss = None

        return Seq2SeqLMOutput(
            loss=masked_lm_loss,
            logits=lm_logits,
            past_key_values=outputs.past_key_values,
            decoder_hidden_states=outputs.decoder_hidden_states,
            decoder_attentions=outputs.decoder_attentions,
            cross_attentions=outputs.cross_attentions,
            encoder_last_hidden_state=outputs.encoder_last_hidden_state,
            encoder_hidden_states=outputs.encoder_hidden_states,
            encoder_attentions=outputs.encoder_attentions,
        )

    def prepare_inputs_for_generation(
        self,
        decoder_input_ids: torch.LongTensor,
        past: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        decoder_head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        use_cache: Optional[bool] = None,
        encoder_outputs: Optional[Union[Tuple[torch.Tensor], BaseModelOutput]] = None,
        **kwargs,
    ) -> Dict:
        # cut decoder_input_ids if past is used
        if past is not None:
            decoder_input_ids = decoder_input_ids[:, -1:]

        return {
            "input_ids": None,  # encoder_outputs is defined. input_ids not needed
            "encoder_outputs": encoder_outputs,
            "past_key_values": past,
            "decoder_input_ids": decoder_input_ids,
            "attention_mask": attention_mask,
            "head_mask": head_mask,
            "decoder_head_mask": decoder_head_mask,
            "cross_attn_head_mask": cross_attn_head_mask,
            "use_cache": use_cache,  # change this to avoid caching (presumably for debugging)
        }

    def adjust_logits_during_generation(self, logits, cur_len):
        logits[:, self.config.pad_token_id] = float("-inf")  # never predict pad token.
        return logits

    @staticmethod
    def _reorder_cache(past, beam_idx):
        reordered_past = ()
        for layer_past in past:
            # cached cross_attention states don't have to be reordered -> they are always the same
            reordered_past += (
                tuple(past_state.index_select(0, beam_idx) for past_state in layer_past[:2]) + layer_past[2:],
            )
        return reordered_past


from transformers.configuration_utils import PretrainedConfig

class MarianOTConfig(PretrainedConfig):
    model_type = "marian_ot"
    keys_to_ignore_at_inference = ["past_key_values"]
    attribute_map = {"num_attention_heads": "encoder_attention_heads", "hidden_size": "d_model"}

    def __init__(
        self,
        vocab_size=58101,
        decoder_vocab_size=None,
        max_position_embeddings=1024,
        encoder_layers=12,
        encoder_ffn_dim=4096,
        encoder_attention_heads=16,
        decoder_layers=12,
        decoder_ffn_dim=4096,
        decoder_attention_heads=16,
        encoder_layerdrop=0.0,
        decoder_layerdrop=0.0,
        use_cache=True,
        is_encoder_decoder=True,
        activation_function="gelu",
        d_model=1024,
        dropout=0.1,
        attention_dropout=0.0,
        activation_dropout=0.0,
        init_std=0.02,
        decoder_start_token_id=58100,
        classifier_dropout=0.0,
        scale_embedding=False,
        pad_token_id=58100,
        eos_token_id=0,
        forced_eos_token_id=0,
        share_encoder_decoder_embeddings=True,
        **kwargs
    ):
        self.vocab_size = vocab_size
        self.decoder_vocab_size = decoder_vocab_size or vocab_size
        self.max_position_embeddings = max_position_embeddings
        self.d_model = d_model
        self.encoder_ffn_dim = encoder_ffn_dim
        self.encoder_layers = encoder_layers
        self.encoder_attention_heads = encoder_attention_heads
        self.decoder_ffn_dim = decoder_ffn_dim
        self.decoder_layers = decoder_layers
        self.decoder_attention_heads = decoder_attention_heads
        self.dropout = dropout
        self.attention_dropout = attention_dropout
        self.activation_dropout = activation_dropout
        self.activation_function = activation_function
        self.init_std = init_std
        self.encoder_layerdrop = encoder_layerdrop
        self.decoder_layerdrop = decoder_layerdrop
        self.classifier_dropout = classifier_dropout
        self.use_cache = use_cache
        self.num_hidden_layers = encoder_layers
        self.scale_embedding = scale_embedding  # scale factor will be sqrt(d_model) if True
        self.share_encoder_decoder_embeddings = share_encoder_decoder_embeddings
        super().__init__(
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
            is_encoder_decoder=is_encoder_decoder,
            decoder_start_token_id=decoder_start_token_id,
            forced_eos_token_id=forced_eos_token_id,
            **kwargs,
        )


import torch
import torch.nn as nn
import pytorch_lightning as pl
from typing import List
from transformers import MarianTokenizer
from marian import MarianOT, MarianOTConfig
import evaluate
import numpy as np

class LitMarianOT(pl.LightningModule):
    def __init__(self, pretrained_ck: str):
        super(LitMarianOT, self).__init__()
        config = MarianOTConfig.from_pretrained(pretrained_ck)
        self.model = MarianOT(config)
        self.tokenizer = MarianTokenizer.from_pretrained(pretrained_ck)
        self.vocab_size = self.tokenizer.vocab_size
        self.loss = nn.CrossEntropyLoss()
        self.metric_tgt = evaluate.load('sacrebleu')
        self.metric_tsl = evaluate.load('sacrebleu')
        self.save_hyperparameters()

    def export_model(self, path):
        self.model.save_pretrained(path)

    def __postprocess(self, predictions, labels):
        predictions = predictions.cpu().numpy()
        labels = labels.cpu().numpy()

        decoded_preds = self.tokenizer.batch_decode(predictions, skip_special_tokens=True)

        # Replace -100 in the labels as we can't decode them.
        labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)
        decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=True)

        # Some simple post-processing
        decoded_preds = [pred.strip() for pred in decoded_preds]
        decoded_labels = [[label.strip()] for label in decoded_labels]
        return decoded_preds, decoded_labels

    def training_step(self, batch, batch_idx):
        tsl_labels = batch.pop('tsl_labels').reshape(-1).long()
        tgt_labels = batch.pop('tgt_labels').reshape(-1).long()

        res = self.model.forward_pass(**batch)
        tgt_logits = res.tgt.logits.reshape(-1, self.vocab_size)
        tsl_logits = res.tsl.logits.reshape(-1, self.vocab_size)
        tgt_loss = self.loss(tgt_logits, tgt_labels)
        tsl_loss = self.loss(tsl_logits, tsl_labels)
        loss = tgt_loss + tsl_loss
        self.log("train/loss", loss, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        tsl_labels = batch.pop('tsl_labels')
        preds_tsl = self.model.generate(batch['input_ids'], max_length=128, num_beams=4)
        decoded_preds, decoded_labels = self.__postprocess(preds_tsl, tsl_labels)
        self.metric_tsl.add_batch(predictions=decoded_preds, references=decoded_labels)

        tgt_labels = batch.pop('tgt_labels')
        with self.model.translate_self():
            preds_tgt = self.model.generate(batch['input_ids'], max_length=128, num_beams=4)
            decoded_preds, decoded_labels = self.__postprocess(preds_tgt, tgt_labels)
            self.metric_tgt.add_batch(predictions=decoded_preds, references=decoded_labels)

    def validation_epoch_end(self, outputs):
        results = self.metric_tgt.compute()
        self.log('valid/bleu_tgt', results['score'], on_epoch=True, on_step=False, sync_dist=True)
        results = self.metric_tsl.compute()
        self.log('valid/bleu_tsl', results['score'], on_epoch=True, on_step=False, sync_dist=True)
        
    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(self.parameters(), lr=3e-5)
        return optimizer


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='compromise-marian',
    packages=['marian'],
    version='0.1.1',
    description='Marian model but with two decoders',
    author='Le Minh Khoi',
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    py_modules=["compromise-marian"],
    install_requires=[]
)


from datasets import load_dataset
from torch.utils.data import DataLoader
from collections.abc import Mapping
import torch
from torch import Tensor
from typing import List, Tuple
from transformers import MarianTokenizer

class MarianOTDataLoader:
    def __init__(self, pretrained_ck: str, max_length: int):
        dataset = load_dataset('kde4', lang1='en', lang2='fr').remove_columns('id')
        dataset = dataset['train'].train_test_split(train_size=0.9, seed=42)
        dataset['valid'] = dataset.pop('test')
        self.dataset = dataset
        self.tokenizer = MarianTokenizer.from_pretrained(pretrained_ck)
        self.max_length = max_length

    def __shift_tokens_right(self, input_ids: torch.Tensor, pad_token_id: int, decoder_start_token_id: int):
        """
        Shift input ids one token to the right.
        """
        shifted_input_ids = input_ids.new_zeros(input_ids.shape)
        shifted_input_ids[:, 1:] = input_ids[:, :-1].clone()
        shifted_input_ids[:, 0] = decoder_start_token_id

        if pad_token_id is None:
            raise ValueError("self.model.config.pad_token_id has to be defined.")
        # replace possible -100 values in labels by `pad_token_id`
        shifted_input_ids.masked_fill_(shifted_input_ids == -100, pad_token_id)

        return shifted_input_ids

    def __collate_fn(self, examples):
        if isinstance(examples, (list, tuple)) and isinstance(examples[0], Mapping):
            encoded_inputs = {key: [example[key] for example in examples] for key in examples[0].keys()}
        else:
            encoded_inputs = examples

        tok = self.tokenizer(
            encoded_inputs['en'],
            text_target=encoded_inputs['fr'],
            max_length=self.max_length,
            padding='max_length',
            truncation=True,
            return_tensors='pt')
        _tok = self.tokenizer(
            encoded_inputs['en'],
            text_target=encoded_inputs['en'],
            max_length=self.max_length,
            padding='max_length',
            truncation=True,
            return_tensors='pt')
        tok['tsl_labels'] = tok.pop('labels')
        tok['tsl_decoder_input_ids'] = self.__shift_tokens_right(tok['tsl_labels'], 59513, 59513)
        tok['tgt_labels'] = _tok.pop('labels')
        tok['tgt_decoder_input_ids'] = self.__shift_tokens_right(tok['tgt_labels'], 59513, 59513)
        
        return tok

    def get_dataloader(self, batch_size:int=16, types: List[str] = ["train", "valid"]):
        res = []
        for type in types:
            res.append(
                DataLoader(self.dataset[type]['translation'], batch_size=batch_size, collate_fn=self.__collate_fn, num_workers=24)
            )
        return res
